from wsknn.weighting.item_weighting import inv_pos_item_score, linear_item_score, log_item_score, quadratic_item_score
from wsknn.weighting.session_weighting import linear_session_score, log_session_score, quadratic_session_score


# Items

def weight_item_score(fn_name: str, element_pos: int) -> float:
    """Function calculates item weight based on its position in a session.

    Parameters
    ----------
    fn_name : str
              Function used for item weighting. Available options: 'linear', 'div', 'log', 'quadratic'.

    element_pos : int
                  Position of an element in a sequence.

    Returns
    -------
    float
        Item weight.
    """

    weighting = {
        'linear': linear_item_score,
        'inv': inv_pos_item_score,
        'log': log_item_score,
        'quadratic': quadratic_item_score
    }

    if fn_name in weighting:
        return weighting[fn_name](element_pos)
    else:
        return 1


# Sessions

def weight_session_items(fn_name: str, element_pos: int, length: int) -> float:
    """Function weights session items by specific fn.

    Parameters
    ----------
    fn_name : str
              Function used for session weighting. Available options: 'linear', 'log', 'quadratic'.

    element_pos : int
                  Position of an element in a sequence.

    length : int
             Overall session length.

    Returns
    -------
    float
        Session's item weight.
    """

    weighting = {
        'linear': linear_session_score,
        'log': log_session_score,
        'quadratic': quadratic_session_score
    }

    return weighting[fn_name](element_pos, length)



import pickle


def load_pickled(filename: str) -> dict:
    """Method loads pickled items / sessions object.

    Parameters
    ----------
    filename : str

    Returns
    -------
    pickled_object : dict
    """
    with open(filename, 'rb') as stored_data:
        pickled_object = pickle.load(stored_data)
    return pickled_object




from wsknn.model.wsknn import WSKNN


def predict(model: WSKNN,
            sessions: dict,
            settings_dict=None) -> dict:
    """
    Function is an alias to predict method of VSKNN model.

    Parameters
    ----------
    model : WSKNN
            Fitted VSKNN model.

    sessions : dict
               Sequence of items for recommendation. As a dict, it is a user id (key) and products and their timestamps
               (values).

    settings_dict : None or dict
                    If given them it overwrites initial settings of a model. Available fields are:
                    - number_of_recommendations : int
                    - number_of_closest_neighbors : int
                    - session_sampling_strategy: str, available options: 'random', 'recent', 'common_items'
                    - possible_neighbors_sample_size : int
                    - weighting_strategy: str, available options: 'linear', 'log', 'quadratic'
                    - rank_strategy: str, available options: 'linear', 'inv', 'log', 'quadratic'

    Returns
    -------
    predicted : dict or list
                List of tuples (Item, Weight) of the length given by number_of_recommendations parameter in
                a descending order; or the same

    Raises
    ------
    ValueError
        Model not fitted.
    """
    if model.item_session_map is None or model.session_item_map is None:
        raise ValueError('Given model does not have an item map and a session map. Fit those before prediction')

    if settings_dict is None:
        predicted = model.predict(sessions)
    else:
        predicted = model.predict(sessions, **settings_dict)
    return predicted



from wsknn.model.wsknn import WSKNN
from wsknn.utils.errors import TooShortSessionException


def mean_reciprocal_rank(sessions: dict, trained_model: WSKNN, r_number=0, skip_short_sessions=True) -> float:
    """
    Function calculates mean reciprocal rank of a top r_number recommendations.
    Given session must be longer than r_number.

    Parameters
    ----------
    sessions : dict
               {'session_id': [
                   [ sequence_of_items ],
                   [ sequence_of_timestamps ],
                   [ [OPTIONAL] sequence_of_event_type ]
               ]}

    trained_model : WSKNN
                    Trained VSKNN model.

    r_number : int, default=0
               Number of top recommendations. Session must have n+1 items minimum to calculate MRR. Default is 0 and
               when it is set, then r_number is equal to number of recommendations from a trained model. If r_number
               is greater than the number of recommendations then the latter is adjusted to it.

    skip_short_sessions : bool, default=True
                          Should algorithm skip short sessions when calculates MRR or should raise an error?

    Returns
    -------
    : float
        Mean Reciprocal Rank: The average score of MRR per n sessions.
    """
    mrr = 0
    denom = 0

    # Check r
    if r_number == 0:
        r_number = trained_model.n_of_recommendations

    if r_number > trained_model.n_of_recommendations:
        trained_model.n_of_recommendations = r_number

    for k, session in sessions.items():

        s_length = len(session[0])

        if s_length <= r_number:
            if skip_short_sessions:
                pass
            else:
                raise TooShortSessionException(s_length, r_number)
        else:
            test_session = [x[:-r_number] for x in session]
            eval_items = session[0][-r_number:]
            predictions = trained_model.predict(
                {k: test_session}
            )

            # Get rank
            rank = 0
            for idx, prod in enumerate(predictions[k]):
                if prod[0] in eval_items:
                    rank = 1 / idx
                    break

            mrr = mrr + rank
            denom = denom + 1

    mrr = mrr / denom
    return mrr



from typing import Union


def check_data_dimension(record: list, dimensions: int) -> bool:
    """Method checks if dimension of a record is equal to dimensions parameter.

    Parameters
    ----------
    record : list
             Nested iterable.

    dimensions : int
                 Number of nested dimensions.

    Returns
    -------
    : bool
        True - dimensions are equal to number of the nested sequences. False otherwise.
    """
    return len(record) == dimensions


class InvalidDimensionsError(Exception):
    """Exception raised when dimension of input session-items map or input item-sessions map is not valid.

    Attributes
    ----------
    set_type : str
               Item-sessions map or session-items map.

    expected_dimensions : int
    """

    def __init__(self, set_type: str, expected_dimensions: Union[int, list]):
        self.set_type = set_type
        self.expected_dimensions = f'{expected_dimensions}'

    def __str__(self):
        msg = f'Expected number(s) of nested sequences for {self.set_type} is {self.expected_dimensions}'
        return msg


class InvalidTimestampError(Exception):
    """Exception raised when timestamp is given in a different format than int.
    """

    def __init__(self, timestamp_sample):
        self.timestamp_type = type(timestamp_sample)

    def __str__(self):
        msg = f'Expected timestamp type should be int, got {self.timestamp_type} instead.'
        return msg


class TooShortSessionException(Exception):
    """Exception raised by Mean Reciprocal Rank metrics if session length is shorter than the
    number of top n recommendations."""

    def __init__(self, session_length: int, top_n: int):
        self.s_length = session_length
        self.topn = top_n
        self.msg = f'Given session length {self.s_length} is shorter than the top n recommendations {self.topn}.\n' \
                   f'It is not possible to make a valid evaluation.'

    def __str__(self):
        return self.msg



from wsknn.utils.calc import is_there_any_common_element, weight_set_pair


def test_weight_set_pair():
    set_a = {1, 3, 5, 7}
    set_b = {0}
    set_c = {1, 0, 9, 3, 8}

    weights_ab = dict()
    weights_ac = {1: 3,
                  3: 12}
    weights_bc = {0: 1.1}

    expected_ab = 0
    expected_ac = 7.5
    expected_bc = 1.1

    assert weight_set_pair(set_a, set_b, weights_ab) == expected_ab
    assert weight_set_pair(set_a, set_c, weights_ac) == expected_ac
    assert weight_set_pair(set_b, set_c, weights_bc) == expected_bc


def test_is_there_any_common_element():
    set_a = {1, 3, 5, 7}
    set_b = {0}
    set_c = {1, 0, 9, 3, 8}

    expected_ab = 0
    expected_bc = 1

    assert is_there_any_common_element(set_a, set_b) == expected_ab
    assert is_there_any_common_element(set_b, set_c) == expected_bc



def weight_set_pair(first: set, second: set, mapped_items_weights: dict) -> float:
    """ Function calculates weighted average of the common items from two sessions based on the dict with items and
        their weights.

    Parameters
    ----------
    first : set
            unique items from the set A

    second : set
             unique items from the set B

    mappped_items_weights : dict
                            {item_id : weight}

    Returns
    -------
    float
        Estimated weight between two item sets. Always 0 or positive.
    """
    if len(mapped_items_weights) >= 1:
        common_items = first & second
        weights_sum = 0
        for i in common_items:
            weights_sum += mapped_items_weights[i]
        result = weights_sum / len(mapped_items_weights)
        return result
    return 0


def is_there_any_common_element(first: set, second: set) -> int:
    """Function checks if there are common elements in two sets and if there is any then it returns 1, else 0.

    Parameters
    ----------
    first : set
            unique items from the set A

    second : set
             unique items from the set B

    Returns
    -------
    bool
        True if there is any common element between sets A and B; 0 otherwise
    """

    return int(bool(first & second))



import pytest
from wsknn.model.wsknn import WSKNN
from wsknn.utils.errors import InvalidDimensionsError, InvalidTimestampError


def test_vsknn_invalid_dimensions_exception():

    sessions = {
        'A': [1, 2, 3, 4],
        'B': [4, 3, 4, 3, 3],
        'C': [10, 2, 1, 1, 1]
    }

    items = {
        '0': 0
    }

    model = WSKNN()

    with pytest.raises(InvalidDimensionsError):
        model.fit(sessions=sessions, items=items)


def test_vsknn_invalid_type_exception():
    sessions = {
        'A': [1, 2, 3],
        'B': [4, 3],
        'C': [10, 2, 1]
    }

    items = {
        '0': 0
    }

    model = WSKNN()

    with pytest.raises(TypeError):
        model.fit(sessions=sessions, items=items)

def test_vsknn_timestamp_type_exception():
    sessions = {
        'A': [[1, 2, 3],
              [10, 20, 30]]
    }

    items = {
        '0': [
            ['A'],
            ['2022-02-22']
        ]

    }

    model = WSKNN()

    with pytest.raises(InvalidTimestampError):
        model.fit(sessions=sessions, items=items)

def test_vsknn_flow1():
    # Type of sessions is str, type of items is int
    sessions = {
        'a': [
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5]
        ],
        'b': [
            [2, 3, 4, 5],
            [10, 11, 12, 13]
        ]
    }

    items = {
        1: [['a'], [1]],
        2: [['a', 'b'], [2]],
        3: [['a', 'b'], [3]],
        4: [['a', 'b'], [4]],
        5: [['a', 'b'], [5]]
    }

    usessions = {'user1': [[1], [100]],
                'user2': [[2, 3], [200, 300]]}

    model = WSKNN()
    model.fit(sessions, items)

    predictions = model.predict(usessions, number_of_recommendations=10, number_of_closest_neighbors=10)
    expected_predictions = {
        'user1': [(2, 2.0), (3, 2.0), (4, 2.0), (5, 2.0)],
        'user2': [(4, 2.5), (5, 2.5), (1, 1.25)]
    }

    assert predictions == expected_predictions


def test_vsknn_flow2():
    # Type of sessions is int, type of items is str
    sessions = {
        0: [
            ['1', '2', '3', '4', '5'],
            [1, 2, 3, 4, 5]
        ],
        1: [
            ['2', '3', '4', '5'],
            [10, 11, 12, 13]
        ]
    }

    items = {
        '1': [[0], [1]],
        '2': [[0, 1], [2]],
        '3': [[0, 1], [3]],
        '4': [[0, 1], [4]],
        '5': [[0, 1], [5]]
    }

    usessions = {'user1': [['1'], [100]],
                'user2': [['2', '3'], [200, 300]]}

    model = WSKNN()
    model.fit(sessions, items)

    predictions = model.predict(usessions, number_of_recommendations=10, number_of_closest_neighbors=10)
    expected_predictions = {
        'user1': [('2', 2.0), ('3', 2.0), ('4', 2.0), ('5', 2.0)],
        'user2': [('4', 2.5), ('5', 2.5), ('1', 1.25)]
    }

    assert predictions == expected_predictions



import yaml


def parse_settings(settings_file: str) -> dict:
    """Function parses settings file into dict

    Parameters
    ----------
    settings_file : str
                    File with the model settings, must be in yaml.

    Returns
    -------
    ydict : dict
            Parsed settings used for modeling.
    """

    with open(settings_file, 'r') as fstream:
        ydict = yaml.safe_load(fstream)

    return ydict



from numpy import log10


def linear_session_score(i, length):
    """Newest elements in sequence => largest weights.

    Parameters
    ----------
    i : int
        Element position, i+1 must be less than length.

    length : int
             Length of a sequence.

    Results
    -------
    result : float
             Session rank between 0 and 1.
    """

    result = (i + 1) / length
    return result


def log_session_score(i, length):
    """Newest element in sequence => largest weight.

    Summary
    -------
    Function calculates weights based on the sequence length and a position of element in a sequence.
        Longer sequences create more steps between limits of 0 and 2.31.
        For example:

        a) element position: 1, sequence length: 10   -> weight = 0.97
        b) element position: 1, sequence length: 1000 -> weight = 0.33

    In practice it means that the past elements of the longest sequences have lower weights than the past elements
        of a short sequence. It mimics our short-term memory and fn gives the biggest weights to the newest events and
        past and middle-sequence events are treated as a not relevant, approx.: to 90% of the sequence.

    Parameters
    ----------
    i : int
        Element position, i+1 must be less than length.

    length : int
             Length of a sequence.

    Results
    -------
    result : float
             Session rank between 0 and 1. Normalized from 0:2.31.
    """

    norm_term = 1 / (log10(2.7))  # The largest possible value
    result = 1 / (log10((length - i) + 1.7))

    return result / norm_term


def quadratic_session_score(i, length):
    """Function penalizes old elements in sequence more than the new events. The past elements have much smaller
       weights than the new ones. It follows a quadratic function.

    Parameters
    ----------
    i : int
        Element position, i+1 must be less than length.

    length : int
             Length of a sequence.

    Results
    -------
    result : float
             Session rank between 0 and 1.
    """

    c = i / length
    result = c*c
    return result


from wsknn.model.wsknn import WSKNN


def fit(sessions: dict, items: dict) -> WSKNN:
    """Function fits given session maps and item maps into a wsknn model.

    Parameters
    ----------
    sessions : dict
               sessions = {
                   session_id: (
                       [sequence_of_items],
                       [sequence_of_timestamps],
                       [sequence_of_event_type]
                   )
               }

    items : dict
            items = {
                item_id: (
                    [sequence_of_sessions],
                    [sequence_of_the_first_session_timestamps]
                )
            }

    Returns
    -------
    wsknn : WSKNN

    Examples
    --------
    >>> input_sessions = {
    ...                     'session_x': (
    ...                                      ['a', 'b', 'c'],
    ...                                      [10001, 10002, 10004],
    ...                                      ['view', 'click', 'click']
    ...                                  )
    ...                  }
    >>> input_items = {
    ...                   'a': (
    ...                            ['session_x'],
    ...                            [10001]
    ...                        ),
    ...                   'b': (
    ...                            ['session_x'],
    ...                            [10001]
    ...                        ),
    ...                   'c': (
    ...                            ['session_x'],
    ...                            [10001]
    ...                        ),
    ...               }
    >>> fitted_model = fit(input_sessions, input_items)
    """

    # Initilize VSKNN model
    vsknn = WSKNN()
    vsknn.fit(
        sessions,
        items
    )

    return vsknn



import numpy as np
from wsknn.weighting.weighting import linear_session_score, log_session_score, quadratic_session_score
from wsknn.weighting.weighting import linear_item_score, inv_pos_item_score, log_item_score, quadratic_item_score

import matplotlib.pyplot as plt


def analyse_weighting_output():

    arr10_length = 10
    arr100_length = 100
    arr1000_length = 1000

    arr10 = np.arange(0, arr10_length, 1)
    arr100 = np.arange(0, arr100_length, 1)
    arr1000 = np.arange(0, arr1000_length, 1)

    arrs = [(arr10_length, arr10),
            (arr100_length, arr100),
            (arr1000_length, arr1000)]

    results = {}
    for arr_tpl in arrs:
        length = arr_tpl[0]
        arr = arr_tpl[1]
        lin_res = [linear_session_score(v, length) for v in arr]
        log_res = [log_session_score(v, length) for v in arr]
        quad_res = [quadratic_session_score(v, length) for v in arr]

        results[length] = (
            lin_res, log_res, quad_res
        )

    for fnl in results.keys():
        plt.figure(figsize=(8, 12))
        plt.plot(results[fnl][0])
        plt.plot(results[fnl][1])
        plt.plot(results[fnl][2])
        plt.legend(['linear', 'log', 'quadratic'])
        plt.show()


def analyse_ranking_output():

    arr10_length = 6
    arr100_length = 10
    arr1000_length = 100

    arr10 = np.arange(1, arr10_length, 1)
    arr100 = np.arange(1, arr100_length, 1)
    arr1000 = np.arange(1, arr1000_length, 1)

    arrs = [(arr10_length, arr10),
            (arr100_length, arr100),
            (arr1000_length, arr1000)]

    results = {}
    for arr_tpl in arrs:
        length = arr_tpl[0]
        arr = arr_tpl[1]
        lin_res = [linear_item_score(v) for v in arr]
        inv_res = [inv_pos_item_score(v) for v in arr]
        log_res = [log_item_score(v) for v in arr]
        quad_res = [quadratic_item_score(v) for v in arr]

        results[length] = (
            lin_res, inv_res, log_res, quad_res
        )

    for fnl in results.keys():
        plt.figure(figsize=(12, 8))
        plt.plot(results[fnl][0])
        plt.plot(results[fnl][1])
        plt.plot(results[fnl][2])
        plt.plot(results[fnl][3])
        plt.legend(['linear', 'inverted', 'log', 'quadratic'])
        plt.show()

if __name__ == '__main__':

    # Lab test (1) - uncomment to run
    # analyse_weighting_output()

    # Summary of test:
    #   PAST elements:
    #     - LOG_FN: returns the largest weights for the past elements,
    #     - QUADRATIC_FN: returns the lowest weights for the past elements.
    #   NEAR-PRESENT elements:
    #      - LINEAR_FN: returns the largest values,
    #      - LOG_FN: returns larger values than QUADRATIC_FN for small sequences == ~10 elements.
    #   GENERAL SHAPE:
    #      - LINEAR_FN: constant change of weights,
    #      - LOG_FN: small weights up to the end of sequence, the newest elements gain the most attention, and even
    #                slightly older are not considered to be relevant,
    #      - QUADRATIC_FN: the oldest elements are penalized more than the middle elements in a sequence. Weights are
    #                      always smaller than in the LINEAR_FN but for the largest sequences weights are much
    #                      bigger than for the LOG_FN.

    ###################################################################################################################

    # Lab test (2) - uncomment to run
    analyse_ranking_output()

    # Summary of test:
    #   PAST elements:
    #     - LOG_FN: returns constant weight above 0, in practice it doesn't leave any element based on its position,
    #     - INV_FN: gradually decreases to zero but is always above the zero value, it can be treated as a function
    #               with properties between LOG_FN and LINEAR_FN or QUADRATIC_FN,
    #     - LINEAR_FN and QUADRATIC_FN: large penatlies for the past values, LINEAR_FN >= 10 gives always 0, and
    #                                   QUADRATIC_FN is very close to zero after only a few steps.
    #   NEAR-PRESENT elements:
    #      - LINEAR_FN: returns the largest values,
    #      - LOG_FN: returns the second largest values,
    #      - INV_FN: returns the third largest values,
    #   GENERAL SHAPE:
    #      - LINEAR_FN: constant change of weights, after 10th step all values are set to 0,
    #      - LOG_FN: small weights up to the end of sequence, the newest elements gain the most attention, and even
    #                slightly older are not considered to be relevant,
    #      - QUADRATIC_FN: the oldest elements are penalized more than the middle elements in a sequence. Weights are
    #                      always smaller than in the LINEAR_FN but for the largest sequences weights are much
    #                      bigger than for the LOG_FN.
    #      - INV_FN: Similar to QUADRATIC_FN but values are always larger and they tend to be > than 0.

    print('Test end')



import numpy as np
from wsknn.weighting.session_weighting import linear_session_score, log_session_score, quadratic_session_score
from wsknn.weighting.item_weighting import linear_item_score, inv_pos_item_score, quadratic_item_score, log_item_score


def test_session_log10():
    sequence_size = np.random.randint(0, 10000)
    vals_range = np.arange(0, sequence_size, 1)
    output = [log_session_score(v, sequence_size) for v in vals_range]
    for res in output:
        assert res <= 1
        assert res >= 0


def test_session_lin():
    sequence_size = np.random.randint(0, 10000)
    vals_range = np.arange(0, sequence_size, 1)
    output = [linear_session_score(v, sequence_size) for v in vals_range]
    for res in output:
        assert res <= 1
        assert res >= 0


def test_session_quadratic():
    sequence_size = np.random.randint(0, 10000)
    vals_range = np.arange(0, sequence_size, 1)
    output = [quadratic_session_score(v, sequence_size) for v in vals_range]
    for res in output:
        assert res <= 1
        assert res >= 0


def test_item_linear():
    sequence_size = np.random.randint(0, 10000)
    vals_range = np.arange(1, sequence_size, 1)
    output = [linear_item_score(v) for v in vals_range]
    for res in output:
        assert res <= 1
        assert res >= 0


def test_item_log():
    sequence_size = np.random.randint(0, 10000)
    vals_range = np.arange(1, sequence_size, 1)
    output = [log_item_score(v) for v in vals_range]
    for res in output:
        assert res <= 1
        assert res >= 0


def test_item_inv():
    sequence_size = np.random.randint(0, 10000)
    vals_range = np.arange(1, sequence_size, 1)
    output = [inv_pos_item_score(v) for v in vals_range]
    for res in output:
        assert res <= 1
        assert res >= 0


def test_item_quadratic():
    sequence_size = np.random.randint(0, 10000)
    vals_range = np.arange(1, sequence_size, 1)
    output = [quadratic_item_score(v) for v in vals_range]
    for res in output:
        assert res <= 1
        assert res >= 0



from numpy import log10


def linear_item_score(i):
    """Function weights events based on their position in a sequence. Output is normalized to the range [0:1]. For the
       long sequences of events (index 10+) it returns zero.

    Parameters
    ----------
    i : int
        Item position.

    Returns
    -------
    result : float
             Linear rank.
    """
    result = 1 - (0.1 * i) if i < 10 else 0
    result = result / 0.9
    return result


def inv_pos_item_score(i):
    """Function returns the final rank as a simple inverse of the item position.

    Parameters
    ----------
    i : int
        Item position.

    Returns
    -------
    result : float
             Inverted position rank.
    """
    result = 1 / i
    return result


def log_item_score(i):
    """Function returns rank as a transformed inverted log10 of the sum (position + 1.7).
       Result is normalized to the range (0:1].

    Parameters
    ----------
    i : int
        Item position.

    Returns
    -------
    result : float
             Logarithmic rank.
    """
    norm_term = 1 / (log10(2.7))  # The largest possible value
    result = 1 / log10(i + 1.7)
    result = result / norm_term
    return result


def quadratic_item_score(i):
    """Function is similar to inverted and linear functions but weights are decreasing at non-linear rate
       and accelerate with the item position.

    Parameters
    ----------
    i : int
        Item position.

    Returns
    -------
    result : float
             Inverted square position rank.
    """
    result = 1 / (i * i)
    return result



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="wsknn",
    version="0.1",
    author="Szymon Moliński",
    author_email="s.molinski@digitree.pl",
    description="VSKNN model for recommendations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.sare25.com/s.molinski/vsknn-model",
    packages=setuptools.find_packages(
        exclude=['dev']
    ),
    install_requires=[
        'numpy',
        'pyyaml'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



import random
from typing import Iterable

from wsknn.weighting import weight_session_items, weight_item_score
from wsknn.utils.calc import weight_set_pair
from wsknn.utils.errors import check_data_dimension, InvalidDimensionsError, InvalidTimestampError


class WSKNN:
    """Class represent Weighted Session-Based k-nn model.

    Parameters
    ----------
    number_of_recommendations : int, default=5
                                Number of items recommended.

    number_of_neighbors : int, default=10
                          Number of closest sessions to choose the items from.

    sampling_strategy : str, default='common_items'
                        How to filter the initial sample of sessions. Available strategies are:
                        - 'common_items': sample sessions with the same items as the input session,
                        - 'recent': sample the most actual sessions,
                        - 'random': get a random sample of sessions.

    sample_size : int, default=1000
                  How many sessions from the model are sampled to make a recommendation.

    weighting_func : str, default='linear'
                     The similarity measurement between sessions. Available options: 'linear', 'log' and 'quadratic'.

    ranking_strategy : str, default='linear'
                       How we calculate an item rank (based on its position in a session sequence). Available options
                       are: 'inv', 'linear', 'log', 'quadratic'.

    Attributes
    ----------
    weighting_functions: list
                         List with weighting functions: ['linear', 'log', 'quadratic'].

    ranking_strategies : list
                         List with ranking strategies: ['linear', 'log', 'quadratic', 'inv'].

    session_item_map : dict
                       sessions = {
                           session_id: (
                               [sequence_of_items],
                               [sequence_of_timestamps],
                               [sequence_of_event_type]
                           )
                       }
                       Dict populated with fit() method.

    item_session_map : dict
                       items = {
                           item_id: (
                               [sequence_of_sessions],
                               [sequence_of_the_first_session_timestamps]
                           )
                       }
                       Dict populated with fit() method.

    n_of_recommendations : int
                           Number of items recommended.

    number_of_closest_neighbors : int

    sampling_strategy : str
                        See sampling_strategy parameter.

    possible_neighbors_sample_size : int
                                     See sample_size parameter.

    weighting_function : str
                         See weighting_func parameter.

    ranking_strategy : str
                       See ranking_strategy parameter.

    Methods
    -------
    fit()

    predict()

    Raises
    ------
    InvalidDimensionsError : wrong number of nested sequences within session-items or item-sessions maps.

    InvalidTimestampError : wrong type of timestamp - int is required.

    TypeError : wrong type of nested structures within session-items or item-sessions maps.

    """

    def __init__(self,
                 number_of_recommendations=5,
                 number_of_neighbors=10,
                 sampling_strategy='common_items',
                 sample_size=1000,
                 weighting_func='linear',
                 ranking_strategy='linear'):

        self.sampling_strategies = ['common_items', 'recent', 'random']
        self.weighting_functions = ['linear', 'log', 'quadratic']
        self.ranking_strategies = ['linear', 'log', 'quadratic', 'inv']

        self.session_item_map = None
        self.item_session_map = None

        self.n_of_recommendations = number_of_recommendations
        self.number_of_closest_neighbors = number_of_neighbors
        self.possible_neighbors_sample_size = sample_size

        self.sampling_strategy = self._is_sampling_strategy_valid(sampling_strategy)
        self.weighting_function = self._is_weighting_function_valid(weighting_func)
        self.ranking_strategy = self._is_ranking_strategy_valid(ranking_strategy)

    # Core methods

    def fit(self, sessions: dict, items: dict):
        """Sets input session-items and item-sessions maps.

        Parameters
        ----------
        sessions : dict
                   sessions = {
                       session_id: (
                           [ sequence_of_items ],
                           [ sequence_of_timestamps ],
                           [ [OPTIONAL] sequence_of_event_type ]
                       )
                   }

        items : dict
                items = {
                    item_id: (
                        [ sequence_of_sessions ],
                        [ sequence_of_the_first_session_timestamps ]
                    )
                }
        """

        # Check input data
        self._check_sessions_input(sessions)
        self._check_items_input(items)

        self.session_item_map = sessions
        self.item_session_map = items

    def predict(self, sessions: dict,
                number_of_recommendations=None,
                number_of_closest_neighbors=None,
                session_sampling_strategy=None,
                possible_neighbors_sample_size=None,
                weighting_strategy=None,
                rank_strategy=None) -> dict:
        """Method predicts n next recommendations from a given session.

        Parameters
        ----------
        sessions : dict
                   User ID (key) and sequence of viewed products and their timestamps (values).

        number_of_recommendations : int or None, default=None
                                    Resets number of recommendations.

        number_of_closest_neighbors : int or None, default=None
                                      Resets number of closest neighbors.

        session_sampling_strategy : str or None, default=None
                                    How to filter the initial sample of sessions. Available strategies are:
                                        - 'common_items': sample sessions with the same items as the input session,
                                        - 'recent': sample the most actual sessions,
                                        - 'random': get a random sample of sessions.
                                        
        possible_neighbors_sample_size : int or None, default=None
                                         How many sessions from the model are sampled to make a recommendation. If not
                                         set then it is the number given during the class initilization.

        weighting_strategy : str or None, default=None
                             The similarity measurement between sessions. Available options: 'linear', 'log'
                             and 'quadratic'. If not set then weighting strategy is taken from the parameter set
                             during the class initialization.

        rank_strategy : str or None, default=None
                        How we calculate an item rank (based on its position in a session sequence). Available options
                        are: 'inv', 'linear', 'log', 'quadratic'. If not set then model uses rank strategy given
                        during the initilization.

        Returns
        -------
        : dict
          ranked items in descending order
          - {user ID: [[item, rank] ...]}
        """

        # Set class consts
        self._reset_model_params(number_of_recommendations,
                                 number_of_closest_neighbors,
                                 session_sampling_strategy,
                                 possible_neighbors_sample_size,
                                 weighting_strategy,
                                 rank_strategy)

        output_ranks = dict()
        for _key, session in sessions.items():
            neighbors = self._nearest_neighbors(session)
            ranked_items = self._rank_items(neighbors, session)
            ranked_items.sort(key=lambda x: x[1], reverse=True)
            recommendations = ranked_items[:self.n_of_recommendations]
            output_ranks[_key] = recommendations
        return output_ranks

    # Settings
    @staticmethod
    def _check_sessions_input(sessions: dict):
        """Check if sessions have required dimensions.

        Parameters
        ----------
        sessions : dict
                   sessions = {
                       session_id: (
                           [ sequence_of_items ],
                           [ sequence_of_timestamps ],
                           [ [OPTIONAL] sequence_of_event_type ]
                       )
                   }

        Raises
        ------
        InvalidDimensionsError
            Number of nested sequences per record is different than 2 or 3.

        InvalidTimestampError
            Wrong type of timestamp - int is required.

        TypeError
            Wrong type of a given structure.
        """

        # Get sample record
        sample_key = random.choice(list(sessions.keys()))

        sample_rec = sessions[sample_key]

        # Check dimensions
        test_dims = check_data_dimension(sample_rec, 2) or check_data_dimension(sample_rec, 3)

        if not test_dims:
            raise InvalidDimensionsError('Session-items map', [2, 3])

        # Check type
        for subrec in sample_rec:
            if not isinstance(subrec, Iterable):
                msg = f'Expected iterable as a part of session-items map record, got {type(subrec)} instead.'
                raise TypeError(msg)

        # Check timestamp type
        for tstamp in sample_rec[1]:
            if not isinstance(tstamp, int):
                raise InvalidTimestampError(tstamp)

    @staticmethod
    def _check_items_input(items: dict):
        """Check if sessions have required dimensions.

        Parameters
        ----------
        items : dict
                items = {
                    item_id: (
                        [ sequence_of_sessions ],
                        [ sequence_of_the_first_session_timestamps ]
                    )
                }

        Raises
        ------
        InvalidDimensionsError
            Number of nested sequences per record is different than 2.

        InvalidTimestampError
            Wrong type of timestamp - int is required.

        TypeError
            Wrong type of a given structure.

        """

        # Get sample record
        sample_key = random.choice(list(items.keys()))

        sample_rec = items[sample_key]

        # Check dimensions
        test_dims = check_data_dimension(sample_rec, 2)

        if not test_dims:
            raise InvalidDimensionsError('Item-sessions map', 2)

        # Check type
        for subrec in sample_rec:
            if not isinstance(subrec, Iterable):
                msg = f'Expected iterable as a part of item-sessions map record, got {type(subrec)} instead.'
                raise TypeError(msg)

        # Check timestamp type
        for tstamp in sample_rec[1]:
            if not isinstance(tstamp, int):
                raise InvalidTimestampError(tstamp)

    def _is_sampling_strategy_valid(self, sampling_strategy):
        """Check sampling strategy.

        Parameters
        ----------
        sampling_strategy : str

        Raises
        ------
        KeyError
            Strategy not in list defined in self.sampling_strategies

        Returns
        -------
        sampling_strategy : str
        """
        if sampling_strategy in self.sampling_strategies:
            return sampling_strategy
        else:
            msg = f"Given sampling strategy {sampling_strategy} not implemented in the package. " \
                  f"Use one of {self.sampling_strategies} instead"
            raise KeyError(msg)

    def _is_ranking_strategy_valid(self, strategy: str) -> str:
        """Check if ranking strategy is valid.

        Parameters
        ----------
        strategy : str

        Raises
        ------
        KeyError
            Strategy not in list ['linear', 'log', 'quadratic', 'inv']

        Returns
        strategy : str
        """
        if strategy in self.ranking_strategies:
            return strategy
        else:
            msg = f"Given ranking strategy {strategy} not implemented in the package. " \
                  f"Use 'linear', 'log', 'quadratic', 'inv' instead"
            raise KeyError(msg)

    def _is_weighting_function_valid(self, wfunc: str) -> str:
        """Check if weighting function is valid.

        Parameters
        ----------
        wfunc : str

        Raises
        ------
        KeyError
            Weighting function not in list ['linear', 'log', 'quadratic']

        Returns
        wfunc : str
        """
        if wfunc in self.weighting_functions:
            return wfunc
        else:
            msg = f"Given weighting function {wfunc} is not implemented in the package. " \
                  f"Use 'linear', 'log', 'quadratic', instead"
            raise KeyError(msg)

    def _reset_model_params(self,
                            number_of_recommendations,
                            number_of_closest_neighbors,
                            session_sampling_strategy,
                            possible_neighbors_sample_size,
                            weighting_strategy,
                            rank_strategy):
        """Methods resets and maps new model parameters.

        Parameters
        ----------
        number_of_recommendations : int

        number_of_closest_neighbors : int

        session_sampling_strategy : str

        possible_neighbors_sample_size : int

        weighting_strategy : str

        rank_strategy : str

        Raises
        -------
        TypeError
            Wrong type of an input parameter.

        KeyError
            Wrong name of sampling strategy, ranking strategy or weighting function.

        """
        if number_of_recommendations is not None:
            if isinstance(number_of_recommendations, int):
                self._set_n_of_recs(number_of_recommendations)
            else:
                raise TypeError(f'Number of output recommendations should be an integer, '
                                f'got {type(number_of_recommendations)} instead')

        if number_of_closest_neighbors is not None:
            if isinstance(number_of_closest_neighbors, int):
                self._set_number_of_closest_neighbors(number_of_closest_neighbors)
            else:
                raise TypeError(f'Number of closest neighbors should be an integer, '
                                f'got {type(number_of_closest_neighbors)} instead')

        if session_sampling_strategy is not None:
            if isinstance(session_sampling_strategy, str):
                session_sampling_strategy = self._is_sampling_strategy_valid(session_sampling_strategy)
                self._set_sampling_strategy(session_sampling_strategy)
            else:
                raise TypeError(f'Defined sampling strategy should be a string. '
                                f'Got {type(session_sampling_strategy)} instead')

        if possible_neighbors_sample_size is not None:
            if isinstance(possible_neighbors_sample_size, int):
                self._set_possible_neighbors_sample_size(possible_neighbors_sample_size)
            else:
                raise TypeError(f'Number of possible neighbors should be an integer, '
                                f'got {type(possible_neighbors_sample_size)} instead')

        if weighting_strategy is not None:
            if isinstance(weighting_strategy, str):
                weighting_strategy = self._is_weighting_function_valid(weighting_strategy)
                self._set_weighting_strategy(weighting_strategy)
            else:
                raise TypeError(f'Defined weighting function should be a string. '
                                f'Got {type(weighting_strategy)} instead')

        if rank_strategy is not None:
            if isinstance(rank_strategy, str):
                rank_strategy = self._is_ranking_strategy_valid(rank_strategy)
                self._set_ranking_strategy(rank_strategy)
            else:
                raise TypeError(f'Defined ranking function should be a string. '
                                f'Got {type(rank_strategy)} instead')

    def _set_n_of_recs(self, n):
        self.n_of_recommendations = n

    def _set_number_of_closest_neighbors(self, n):
        self.number_of_closest_neighbors = n

    def _set_possible_neighbors_sample_size(self, sample_size):
        self.possible_neighbors_sample_size = sample_size

    def _set_sampling_strategy(self, sampling_strategy):
        self.sampling_strategy = sampling_strategy

    def _set_weighting_strategy(self, weighting_func):
        weighting_func = self._is_weighting_function_valid(weighting_func)
        self.weighting_function = weighting_func

    def _set_ranking_strategy(self, rank_strategy):
        self.ranking_strategy = rank_strategy

    # Transform, sample, rank - core

    def _nearest_neighbors(self, session: list) -> list:
        """Method searches for nearest neighbors for a given session.

        Parameters
        ----------
        session : list
                  Session is a sequence of products, events (product view / click) and timestamps of those events:

            session = [
                [sequence_of_items],
                [sequence_of_timestamps],
                [sequence_of_event_type]
            ]

        Returns
        -------
        : list
          n closest sessions, where n - number of closest sessions or length of ranked session if smaller than
          number of closest sessions.
        """
        possible_neighbor_sessions = self._possible_neighbors(session)
        items_sequence = session[0]
        rank = self._calculate_similarity(items_sequence, possible_neighbor_sessions)
        rank.sort(key=lambda x: x[1], reverse=True)
        length = len(rank)
        idx = min(self.number_of_closest_neighbors, length)
        return rank[0:idx]

    def _possible_neighbors(self, session: list) -> list:
        """Get set of possible neighbors based on the item similarity.

        Parameters
        ----------
        session : list
                  Session is a sequence of products, events (product view / click) and timestamps of those events:

            session = [
                [sequence_of_items],
                [sequence_of_timestamps],
                [sequence_of_event_type]
            ]

        Returns
        -------
        : list
          Sample of possible neighbors. Sampling controlled by the sampling_strategy attribute.
        """
        session_items = set(session[0])
        common_sessions = set()
        for s_item in session_items:
            if s_item in self.item_session_map:
                s_item_sessions = set(self.item_session_map[s_item][0])
                common_sessions |= s_item_sessions
        sample_subset = self._sample_possible_neighbors(common_sessions, session)
        return sample_subset

    def _rank_items(self, closest_neighbors: list, session: list) -> list:
        """Function ranks given items to return the best recommendation results.

        Parameters
        ----------
        closest_neighbors : list
                            The closest sessions ranked by similarity to a given session.

        session : list
                  User session.

        Returns
        -------
        : list
          List of rated items in descending order.
        """
        session_items = session[0]
        scores = dict()

        for neighbor in closest_neighbors:
            n_items = self.session_item_map[neighbor[0]][0]
            step = 1
            decay = 1
            for s_item in reversed(session_items):
                if s_item in n_items:
                    decay = weight_item_score(self.ranking_strategy, step)
                    break
                step = step + 1

            for n_item in n_items:
                if n_item in session_items:
                    # We do not want to score items viewed by the user
                    # TODO: condition to control this behavior
                    pass
                else:
                    old_score = scores.get(n_item)
                    new_score = neighbor[1]
                    # TODO: idf weighting
                    new_score = new_score * decay

                    if old_score is not None:
                        new_score = old_score + new_score

                    scores.update({n_item: new_score})

        rank = list()
        for k, v in scores.items():
            rank.append((k, v))
        return rank

    # Transform, sample, rank - additional

    def _calculate_similarity(self, session_items: list, possible_neighbours: list) -> list:
        """Function calculates similarity between sessions based on the items ranking.

        Parameters
        ----------
        session_items : list
                        List of items from the customer session.

        possible_neighbours : list
                              List of sessions from the possible neighbors pool (based on a sampling strategy).

        Returns
        -------
        neighbors : list
                    List of similar sessions to the customer session.
        """

        pos_weights = dict()
        length = len(session_items)

        for idx, item in enumerate(session_items):
            count = idx + 1
            pos_weights[item] = weight_session_items(self.weighting_function, count, length)

        items = set(session_items)
        neighbours = []
        for other in possible_neighbours:
            other_items = set(self.session_item_map[other][0])
            similarity = weight_set_pair(items, other_items, pos_weights)
            neighbours.append([other, similarity])

        return neighbours

    def _sampling_common(self, sessions: set, session: list) -> list:
        """Function gets the most similar sessions based on the number of common elements between sessions.

        Parameters
        ----------
        sessions : set
                   Unique sessions.

        session : list
                  The customer session.

        Returns
        -------
        : list
          List of n possible sessions with the same items as a customer session.
        """
        rank = [(ses, len(set(self.session_item_map[ses][0]) & set(session[0]))) for ses in sessions]
        rank.sort(key=lambda x: x[1])
        result = [x[0] for x in rank]
        sample_size = min(self.possible_neighbors_sample_size, len(sessions))
        return result[:sample_size]

    def _sample_possible_neighbors(self, all_sessions: set, session: list) -> list:
        """Method samples possible neighbors.

        Parameters
        ----------
        all_sessions : set
                       All sessions to sample possible neighbors.

        session : list
                  Customer session.

        Returns
        -------
        : list
          subset of possible neighbors
        """
        # TODO: Sample based on the click / view

        if self.sampling_strategy == 'random':
            return self._sampling_random(all_sessions)
        elif self.sampling_strategy == 'recent':
            return self._sampling_recent(all_sessions)
        elif self.sampling_strategy == 'common_items':
            return self._sampling_common(all_sessions, session)
        else:
            err_msg = f'Defined sampling strategy {self.sampling_strategy} not implemented. Available strategies are:' \
                      f' {self.sampling_strategies}.'
            raise TypeError(err_msg)

    def _sampling_random(self, sessions: set) -> list:
        """Get random sessions from the sessions space. This method is good to estimate model performance or to test it.

        Parameters
        ----------
        sessions : set
                   Unique sessions.

        Returns
        -------
        : list
          Random sample of self.possible_neighbors_sample_size sessions.
        """
        sample_size = min(self.possible_neighbors_sample_size, len(sessions))
        return random.sample(sessions, sample_size)

    def _sampling_recent(self, sessions: set) -> list:
        """Get most recent sessions from the possible neighbors.

        Parameters
        ----------
        sessions : set
                   Unique sessions.

        Returns
        -------
        : list
          Most recent sessions. Sample of size possible_neighbors_sample_size.
        """
        rank = [(sid, self.session_item_map[sid][1]) for sid in sessions]
        rank.sort(key=lambda x: x[1], reverse=True)
        result = [x[0] for x in rank]
        sample_size = min(self.possible_neighbors_sample_size, len(sessions))
        return result[:sample_size]
